/*=============================================================================
    Phoenix v1.2
    Copyright (c) 2004 Angus Leeming

    Use, modification and distribution is subject to the Boost Software
    License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/

#include "stl_containers.hpp"

#include <boost/spirit/phoenix/primitives.hpp>

#include <iostream>
#include <typeinfo>

#include <deque>
#include <list>
#include <map>
#include <vector>
#include <utility>

using std::cerr;

namespace {

std::deque<int> const build_deque();
std::list<int> const build_list();
std::map<int, int> const build_map();
std::multimap<int, int> const build_multimap();
std::vector<int> const build_vector();


template <typename Container>
bool test_assign(Container c)
{
    using phoenix::arg1;
    using phoenix::assign;

    typename Container::size_type count = 2;
    typename Container::const_iterator first = c.begin();
    typename Container::const_iterator second = first;
    typename Container::value_type value = *first;

    assign(arg1, count, value)(c);
    
    std::advance(second, 1);
    if (*first != *second) {
        cerr << "Failed " << typeid(Container).name() << " test_assign 1\n";
        return false;
    }
#if 0
    // Should not --- does not, Yay! --- compile.
    Container const const_c = c;
    assign(const_c, count, value);
#endif

    return true;
}

template <typename Container>
bool test_assign2(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::arg3;
    using phoenix::assign;

    Container c2 = c;
    typename Container::const_iterator first = c2.begin();
    typename Container::const_iterator last = c2.end();
    typename Container::size_type size = c2.size();

    c.clear();
    assign(arg1, arg2, arg3)(c, first, last);
    if (c.size() != size) {
        cerr << "Failed " << typeid(Container).name()
	     << " test_assign2 1\n"
	     << "size == " << c.size() << '\n';
        return false;
    }

#if 0
    // Should not --- does not, Yay! --- compile.
    Container const const_c = c;
    assign(const_c, first, second);
#endif

    return true;
}


template <typename Container>
bool test_at(Container c)
{
    using phoenix::arg1;
    using phoenix::at;

    typename Container::reference r1 = at(arg1, 0)(c);
    if (r1 != c.at(0)) {
        cerr << "Failed " << typeid(Container).name() << " test_at 1\n";
        return false;
    }

    typename Container::const_reference r2 = at(arg1, 0)(c);
    if (r2 != c.at(0)) {
        cerr << "Failed " << typeid(Container).name() << " test_at 2\n";
        return false;
    }

    Container const const_c = c;
#if 0
    // Should not --- does not, Yay! --- compile.
    typename Container::reference r3 = at(arg1, 0)(const_c);
#endif

    typename Container::const_reference r4 = at(arg1, 0)(const_c);
    if (r4 != c.at(0)) {
        cerr << "Failed " << typeid(Container).name() << " test_at 4\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_back(Container c)
{
    using phoenix::arg1;
    using phoenix::back;

    typename Container::reference r1 = back(arg1)(c);
    if (r1 != c.back()) {
        cerr << "Failed " << typeid(Container).name() << " test_back 1\n";
        return false;
    }
    typename Container::const_reference r2 = back(arg1)(c);
    if (r2 != c.back()) {
        cerr << "Failed " << typeid(Container).name() << " test_back 2\n";
        return false;
    }

    Container const const_c = c;
#if 0
    // Should not --- does not, Yay! --- compile.
    typename Container::reference r3 = back(arg1)(const_c);
#endif

    typename Container::const_reference r4 = back(arg1)(const_c);
    if (r4 != c.back()) {
        cerr << "Failed " << typeid(Container).name() << " test_back 4\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_begin(Container c)
{
    using phoenix::arg1;
    using phoenix::begin;

    typename Container::iterator it1 = begin(arg1)(c);
    if (it1 != c.begin()) {
        cerr << "Failed " << typeid(Container).name() << " test_begin 1\n";
        return false;
    }
    typename Container::const_iterator it2 = begin(arg1)(c);
    if (it2 != c.begin()) {
        cerr << "Failed " << typeid(Container).name() << " test_begin 2\n";
        return false;
    }

    Container const const_c = c;
#if 0
    // Should not --- does not, Yay! --- compile.
    typename Container::iterator it3 = begin(arg1)(const_c);
#endif

    typename Container::const_iterator it4 = begin(arg1)(const_c);
    if (it4 != const_c.begin()) {
        cerr << "Failed " << typeid(Container).name() << " test_begin 4\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_capacity(Container c)
{
    using phoenix::arg1;
    using phoenix::capacity;

    typename Container::size_type s1 = capacity(arg1)(c);
    if (s1 != c.capacity()) {
        cerr << "Failed " << typeid(Container).name() << " test_capacity 1\n";
        return false;
    }

    Container const const_c = c;
    typename Container::size_type s2 = capacity(arg1)(const_c);
    if (s2 != const_c.capacity()) {
        cerr << "Failed " << typeid(Container).name() << " test_capacity 2\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_clear(Container c)
{
    using phoenix::arg1;
    using phoenix::clear;

    clear(arg1)(c);
    if (!c.empty()) {
	    cerr << "Failed " << typeid(Container).name() << " test_clear 1\n";
	    return false;
    }

#if 0
    Container const const_c = c;
    clear(arg1)(const_c);
#endif
    return true;
}


template <typename Container>
bool test_count(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::count;

    typename Container::value_type const value = *c.begin();
    typename Container::key_type const key = value.first;
    typename Container::size_type const counted = count(arg1, arg2)(c, key);
    if (counted != 1) {
        cerr << "Failed " << typeid(Container).name() << " test_count 1\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_empty(Container c)
{
    using phoenix::arg1;
    using phoenix::empty;

    typename Container::size_type s1 = empty(arg1)(c);
    if (s1 != c.empty()) {
        cerr << "Failed " << typeid(Container).name() << " test_empty 1\n";
        return false;
    }

    Container const const_c = c;
    typename Container::size_type s2 = empty(arg1)(const_c);
    if (s2 != const_c.empty()) {
        cerr << "Failed " << typeid(Container).name() << " test_empty 2\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_end(Container c)
{
    using phoenix::arg1;
    using phoenix::end;

    typename Container::iterator it1 = end(arg1)(c);
    if (it1 != c.end()) {
        cerr << "Failed " << typeid(Container).name() << " test_end 1\n";
        return false;
    }
    typename Container::const_iterator it2 = end(arg1)(c);
    if (it2 != c.end()) {
        cerr << "Failed " << typeid(Container).name() << " test_end 2\n";
        return false;
    }

    Container const const_c = c;
#if 0
    // Should not --- does not, Yay! --- compile.
    typename Container::iterator it3 = end(arg1)(const_c);
#endif

    typename Container::const_iterator it4 = end(arg1)(const_c);
    if (it4 != const_c.end()) {
        cerr << "Failed " << typeid(Container).name() << " test_end 4\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_equal_range(Container c)
{
    using phoenix::arg1;
    using phoenix::equal_range;

    typedef typename Container::iterator iterator;
    typedef typename Container::const_iterator const_iterator;
    
    typedef std::pair<iterator, iterator> iterator_pair;
    typedef std::pair<const_iterator, const_iterator> const_iterator_pair;
    
    typename Container::value_type value = *c.begin();
    typename Container::key_type key = value.first;

    iterator_pair ip = equal_range(arg1, arg2)(c, key);
    if (ip.first != c.lower_bound(key) ||
	ip.second != c.upper_bound(key)) {
        cerr << "Failed " << typeid(Container).name()
	     << " test_ equal_range 1\n";
        return false;
    }
	    
    Container const const_c;
    value = *const_c.begin();
    key = value.first;

    const_iterator_pair cip = equal_range(arg1, arg2)(const_c, key);
    if (cip.first != const_c.lower_bound(key) ||
	cip.second != const_c.upper_bound(key)) {
        cerr << "Failed " << typeid(Container).name()
	     << " test_ equal_range 2\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_erase(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::arg3;
    using phoenix::erase;

   Container const const_c = c;

    typename Container::size_type size = c.size();
    typename Container::iterator c_begin = c.begin();
    erase(arg1, arg2)(c, c_begin);
    if (c.size() + 1 != size) {
        cerr << "Failed " << typeid(Container).name() << " test_erase 1\n";
        return false;
    }

    c_begin = c.begin();
    typename Container::iterator c_end = c.end();
    erase(arg1, arg2, arg3)(c, c_begin, c_end);
    if (!c.empty()) {
        cerr << "Failed " << typeid(Container).name() << " test_erase 2\n";
        return false;
    }

#if 0
    erase(arg1, const_c.begin())(const_c);
    erase(arg1, const_c.begin(), const_c.end())(const_c);
#endif
    return true;
}


template <typename Container>
bool test_map_erase(Container c)
{
    if (!test_erase(c))
	    return false;
    
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::erase;

    typename Container::value_type const value = *c.begin();
    typename Container::key_type const key = value.first;
    typename Container::size_type const removed = 
	    erase(arg1, arg2)(c, key);
    if (removed != 1) {
        cerr << "Failed " << typeid(Container).name() << " test_map_erase 1\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_find(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::find;

    Container const const_c;

    typename Container::key_type key = c.begin()->first;

    typename Container::iterator it = find(arg1, arg2)(c, key);
    if (it != c.begin()) {
        cerr << "Failed " << typeid(Container).name()
	     << " test_find 1\n";
        return false;
    }

    key = const_c.begin()->first;

    typename Container::const_iterator cit = find(arg1, arg2)(const_c, key);
    if (cit != const_c.begin()) {
        cerr << "Failed " << typeid(Container).name()
	     << " test_find 2\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_front(Container c)
{
    using phoenix::arg1;
    using phoenix::front;

    typename Container::reference r1 = front(arg1)(c);
    if (r1 != c.front()) {
        cerr << "Failed " << typeid(Container).name() << " test_front 1\n";
        return false;
    }
    typename Container::const_reference r2 = front(arg1)(c);
    if (r2 != c.front()) {
        cerr << "Failed " << typeid(Container).name() << " test_front 2\n";
        return false;
    }

    Container const const_c = c;
#if 0
    // Should not --- does not, Yay! --- compile.
    typename Container::reference r3 = front(arg1)(const_c);
#endif

    typename Container::const_reference r4 = front(arg1)(const_c);
    if (r4 != c.front()) {
        cerr << "Failed " << typeid(Container).name() << " test_front 4\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_get_allocator(Container c)
{
    using phoenix::arg1;
    using phoenix::get_allocator;

    Container const const_c = c;

    typename Container::allocator_type a1 = get_allocator(arg1)(c);
    if (a1 != c.get_allocator()) {
        cerr << "Failed " << typeid(Container).name() << " test_get_allocator 1\n";
        return false;
    }

    typename Container::allocator_type a2 = get_allocator(arg1)(const_c);
    if (a2 != const_c.get_allocator()) {
        cerr << "Failed " << typeid(Container).name() << " test_get_allocator 2\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_insert(Container c)
{
    using phoenix::arg1;
    using phoenix::insert;

    typename Container::value_type const value = *c.begin();
    typename Container::iterator it = insert(arg1, c.begin(), value)(c);
    if (it != c.begin() || *it != *(++it)) {
        cerr << "Failed " << typeid(Container).name() << " test_insert 1\n";
        return false;
    }

    typename Container::size_type size = c.size();
    insert(arg1, c.begin(), 3, value)(c);
    if (c.size() != size + 3) {
        cerr << "Failed " << typeid(Container).name() << " test_insert 2\n";
        return false;
    }

    Container const const_c = c;
    size = c.size();
    insert(arg1, c.begin(), const_c.begin(), const_c.end())(c);
    if (c.size() != 2 * size) {
        cerr << "Failed " << typeid(Container).name() << " test_insert 3\n";
        return false;
    }

    return true;
}


bool test_map_insert(std::map<int, int> c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::arg3;

    typedef std::map<int, int> Map;

    Map::value_type const value = *c.begin();
    Map::iterator c_begin = c.begin();
    // wrapper for
    // iterator insert(iterator where, const value_type& val);
    Map::iterator it =
        phoenix::insert(arg1, arg2, arg3)(c, c_begin, value);

    if (it != c.begin() || *it != *(++it)) {
        cerr << "Failed " << typeid(Map).name() << " test_map_insert 1\n";
        return false;
    }

    // wrapper for
    // pair<iterator, bool> insert(const value_type& val);
    Map::value_type const value2(1400, 2200);
    std::pair<Map::iterator, bool> result =
	    phoenix::insert(arg1, arg2)(c, value2);
    if (!result.second) {
        cerr << "Failed " << typeid(Map).name() << " test_map_insert 2\n";
        return false;
    }

    // wrapper for
    // template<class InIt>
    // void insert(InIt first, InIt last);
    Map const const_c = build_map();
    Map::size_type size = c.size();
    phoenix::insert(arg1, const_c.begin(), const_c.end())(c);
    if (c.size() != size + const_c.size()) {
        cerr << "Failed " << typeid(Map).name() << " test_map_insert 3\n";
        return false;
    }

    return true;
}


bool test_multimap_insert(std::multimap<int, int> c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::arg3;

    typedef std::multimap<int, int> Multimap;

    Multimap::value_type const value = *c.begin();
    Multimap::iterator c_begin = c.begin();
    // wrapper for
    // iterator insert(iterator where, const value_type& val);
    Multimap::iterator it =
        phoenix::insert(arg1, arg2, arg3)(c, c_begin, value);

    if (it != c.begin() || *it != *(++it)) {
        cerr << "Failed " << typeid(Multimap).name()
	     << " test_multimap_insert 1\n";
        return false;
    }

    // wrapper for
    // iterator insert(const value_type& val);
    Multimap::value_type const value2(1400, 2200);
    it = phoenix::insert(arg1, arg2)(c, value2);
    if (it == c.end()) {
        cerr << "Failed " << typeid(Multimap).name()
	     << " test_multimap_insert 2\n";
        return false;
    }

    // wrapper for
    // template<class InIt>
    // void insert(InIt first, InIt last);
    Multimap const const_c = build_multimap();
    Multimap::size_type size = c.size();
    phoenix::insert(arg1, const_c.begin(), const_c.end())(c);
    if (c.size() != size + const_c.size()) {
        cerr << "Failed " << typeid(Multimap).name()
	     << " test_multimap_insert 3\n";
        return false;
    }

    return true;
}


template <typename Container>
bool test_key_comp(Container c)
{
    using phoenix::arg1;
    using phoenix::key_comp;

    typename Container::key_compare comp = key_comp(arg1)(c);

    Container const const_c = c;
    comp = key_comp(arg1)(const_c);

    return true;
}


template <typename Container>
bool test_lower_bound(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::lower_bound;

    typename Container::key_type key = c.begin()->first;

    typename Container::iterator lb1 =
        lower_bound(arg1, arg2)(c, key);

    if (lb1 != c.lower_bound(key)) {
	cerr << "Failed " << typeid(Container).name()
             << " test_lower_bound 1\n";
        return false;
    }

    Container const const_c = c;
    key = const_c.begin()->first;

    typename Container::const_iterator lb2 =
	    lower_bound(arg1, arg2)(const_c, key);

    if (lb2 != const_c.lower_bound(key)) {
	cerr << "Failed " << typeid(Container).name()
             << " test_lower_bound 1\n";
        return false;
    }

    return true;
}


template <typename Container>
bool test_max_size(Container c)
{
    using phoenix::arg1;
    using phoenix::max_size;

    Container const const_c = c;

    typename Container::size_type s1 = max_size(arg1)(c);
    if (s1 != c.max_size()) {
        cerr << "Failed " << typeid(Container).name() << " test_max_size 1\n";
        return false;
    }

    typename Container::size_type s2 = max_size(arg1)(const_c);
    if (s2 != const_c.max_size()) {
        cerr << "Failed " << typeid(Container).name() << " test_max_size 2\n";
        return false;
    }
    return true;
}


struct ReverseSort {
    template <typename Data>
    bool operator()(Data const & a, Data const & b) const
    { return b < a; }
};


template <typename Container>
bool test_merge(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::merge;

    c.sort();
    Container c2 = c;

    typename Container::size_type size = c.size();
    merge(arg1, arg2)(c, c2);
    if (c.size() != 2 * size || !c2.empty()) {
        cerr << "Failed " << typeid(Container).name() << " test_merge 1\n";
        return false;
    }

    size = c.size();
    c2 = c;

    merge(arg1, arg2, ReverseSort())(c, c2);
    if (c.size() != 2 * size || !c2.empty()) {
        cerr << "Failed " << typeid(Container).name() << " test_merge 2\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_pop_back(Container c)
{
    using phoenix::arg1;
    using phoenix::pop_back;

    Container const const_c = c;

    typename Container::size_type size = c.size();

    pop_back(arg1)(c);
    if (c.size() + 1 != size) {
        cerr << "Failed " << typeid(Container).name() << " test_pop_back 1\n";
        return false;
    }

#if 0
    pop_back(arg1)(const_c);
#endif
    return true;
}


template <typename Container>
bool test_pop_front(Container c)
{
    using phoenix::arg1;
    using phoenix::pop_front;

    Container const const_c = c;

    typename Container::size_type size = c.size();

    pop_front(arg1)(c);
    if (c.size() + 1 != size) {
        cerr << "Failed " << typeid(Container).name() << " test_pop_front 1\n";
        return false;
    }
#if 0
    pop_front(arg1)(const_c);
#endif
    return true;
}


template <typename Container>
bool test_push_back(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::push_back;

    Container const const_c = c;

    typename Container::value_type data = *c.begin();
    typename Container::size_type size = c.size();
    push_back(arg1, arg2)(c, data);
    if (c.size() != size + 1) {
        cerr << "Failed " << typeid(Container).name() << " test_push_back 1\n";
        return false;
    }
#if 0
    push_back(arg1, arg2)(const_c, data);
#endif
    return true;
}


template <typename Container>
bool test_push_front(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::push_front;

    Container const const_c = c;

    typename Container::value_type data = *c.begin();
    typename Container::size_type size = c.size();
    push_front(arg1, arg2)(c, data);
    if (c.size() != size + 1) {
        cerr << "Failed " << typeid(Container).name() << " test_push_front 1\n";
        return false;
    }
#if 0
    push_front(arg1, arg2)(const_c, data);
#endif
    return true;
}


template <typename Container>
bool test_rbegin(Container c)
{
    using phoenix::arg1;
    using phoenix::rbegin;

    typename Container::reverse_iterator it1 = rbegin(arg1)(c);
    typename Container::reverse_iterator it1_test = c.rbegin();
    if (it1 != it1_test) {
        cerr << "Failed " << typeid(Container).name() << " test_rbegin 1\n";
        return false;
    }
    typename Container::const_reverse_iterator it2 = rbegin(arg1)(c);
    typename Container::const_reverse_iterator it2_test = c.rbegin();
    if (it2 != it2_test) {
        cerr << "Failed " << typeid(Container).name() << " test_rbegin 2\n";
        return false;
    }

    Container const const_c = c;
#if 0
    // Should not --- does not, Yay! --- compile.
    typename Container::reverse_iterator it3 = rbegin(arg1)(const_c);
#endif

    typename Container::const_reverse_iterator it4 = rbegin(arg1)(const_c);
    it2_test = const_c.rbegin();
    if (it4 != it2_test) {
        cerr << "Failed " << typeid(Container).name() << " test_rbegin 4\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_remove(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::remove;

    typename Container::size_type size = c.size();
    typename Container::value_type value = *c.begin();
    remove(arg1, arg2)(c, value);
    if (c.size() != size - 1) {
        cerr << "Failed " << typeid(Container).name() << " test_remove 1\n";
        return false;
    }
    return true;
}


struct FirstThree {
    FirstThree() : count(0) {}

    template <typename Data>
    bool operator()(Data const &)
    { return count++ < 3; }
private:
    int count;
};


template <typename Container>
bool test_remove_if(Container c)
{
    using phoenix::arg1;
    using phoenix::remove_if;

    typename Container::size_type size = c.size();
    // Remove first three elements
    remove_if(arg1, FirstThree())(c);
    if (c.size() != size - 3) {
        cerr << "Failed " << typeid(Container).name() << " test_remove_if 1\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_rend(Container c)
{
    using phoenix::arg1;
    using phoenix::rend;

    typename Container::reverse_iterator it1 = rend(arg1)(c);
    typename Container::reverse_iterator it1_test = c.rend();
    if (it1 != it1_test) {
        cerr << "Failed " << typeid(Container).name() << " test_rend 1\n";
        return false;
    }
    typename Container::const_reverse_iterator it2 = rend(arg1)(c);
    typename Container::const_reverse_iterator it2_test = c.rend();
    if (it2 != it2_test) {
        cerr << "Failed " << typeid(Container).name() << " test_rend 2\n";
        return false;
    }

    Container const const_c = c;
#if 0
    // Should not --- does not, Yay! --- compile.
    typename Container::reverse_iterator it3 = rend(arg1)(const_c);
#endif

    typename Container::const_reverse_iterator it4 = rend(arg1)(const_c);
    it2_test = const_c.rend();
    if (it4 != it2_test) {
        cerr << "Failed " << typeid(Container).name() << " test_rend 4\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_reserve(Container c)
{
    using phoenix::arg1;
    using phoenix::reserve;

    Container const const_c = c;

    typename Container::size_type count = 2 * c.size();
    reserve(arg1, count)(c);
    if (c.capacity() < count) {
        cerr << "Failed " << typeid(Container).name() << " test_reserve 1\n";
        return false;
    }
#if 0
    reserve(arg1, count)(const_c)(const_c);
#endif
    return true;
}


template <typename Container>
bool test_resize(Container c)
{
    using phoenix::arg1;
    using phoenix::resize;

    Container const const_c = c;

    typename Container::size_type new_size = 2 * c.size();
    resize(arg1, new_size)(c);
    if (c.size() != new_size) {
        cerr << "Failed " << typeid(Container).name() << " test_resize 1\n";
        return false;
    }

    new_size = 2 * c.size();
    typename Container::value_type value = *c.begin();
    resize(arg1, new_size, value)(c);
    if (c.size() != new_size) {
        cerr << "Failed " << typeid(Container).name() << " test_resize 2\n";
        return false;
    }
#if 0
    new_size = 2 * const_c.size();
    resize(arg1, new_size)(const_c);

    new_size = 2 * const_c.size();
    resize(arg1, new_size, value)(const_c);
#endif
    return true;
}


template <typename Container>
bool test_reverse(Container c)
{
    using phoenix::arg1;
    using phoenix::reverse;

    typename Container::value_type data = *(c.end()--);
    reverse(arg1)(c);
    if (data != *(c.end()--)) {
        cerr << "Failed " << typeid(Container).name() << " test_reverse 1\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_size(Container c)
{
    using phoenix::arg1;
    using phoenix::size;

    Container const const_c = c;

    typename Container::size_type s1 = size(arg1)(c);
    if (s1 != c.size()) {
        cerr << "Failed " << typeid(Container).name() << " test_size 1\n";
        return false;
    }

    typename Container::size_type s2 = size(arg1)(const_c);
    if (s2 != const_c.size()) {
        cerr << "Failed " << typeid(Container).name() << " test_size 2\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_sort(Container c)
{
    using phoenix::arg1;
    using phoenix::sort;

    Container copy = c;

    sort(arg1, ReverseSort())(c);
    sort(arg1)(c);
    // Implies that c was sorted to begin with.
    if (c != copy) {
        cerr << "Failed " << typeid(Container).name() << " test_sort 1\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_splice(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::arg3;
    using phoenix::arg4;
    using phoenix::arg5;
    using phoenix::splice;

    typename Container::iterator c_end;
    typename Container::iterator c2_begin;
    typename Container::iterator c2_end;
    typename Container::size_type size = c.size();

    Container const copy = c;
    Container const copy2 = build_list();
    Container c2 = copy2;

    size = c.size();
    c_end = c.end();
    splice(arg1, arg2, arg3)(c, c_end, c2);
    if (c.size() != 2 * size) {
        cerr << "Failed " << typeid(Container).name() << " test_splice 1\n";
        return false;
    }

    c = copy;
    c_end = c.end();
    c2 = copy2;
    c2_begin = c2.begin();
    size = c.size() + 1;
    splice(arg1, arg2, arg3, arg4)(c, c_end, c2, c2_begin);
    if (c.size() != size) {
        cerr << "Failed " << typeid(Container).name() << " test_splice 2\n";
        return false;
    }

    c = copy;
    c_end = c.end();
    c2 = copy2;
    c2_begin = c2.begin();
    c2_end = c2.end();
    size = c.size() + c2.size();
    splice(arg1, arg2, arg3, arg4, arg5)(c, c_end, c2, c2_begin, c2_end);
    if (c.size() != size) {
        cerr << "Failed " << typeid(Container).name() << " test_splice 3\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_swap(Container c1)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::swap;

    Container c2 = c1;

    swap(arg1, arg2)(c1, c2);

#if 0
    Container const const_c1 = c1;
    Container const const_c2 = c1;
    swap(arg1, arg2)(const_c1, const_c2);
#endif
    return true;
}


struct Equal {
    template <typename Data>
    bool operator()(Data const & d1, Data const & d2) const
    { return d1 == d2; }
};


template <typename Container>
bool test_unique(Container c)
{
    using phoenix::arg1;
    using phoenix::unique;

    typename Container::size_type size = c.size();
    typename Container::value_type val = *c.begin();
    c.insert(c.begin(), val);

    unique(arg1)(c);
    if (c.size() != size) {
        cerr << "Failed " << typeid(Container).name() << " test_unique 1\n";
        return false;
    }
    
    c.insert(c.begin(), val);
    unique(arg1, Equal())(c);
    if (c.size() != size) {
        cerr << "Failed " << typeid(Container).name() << " test_unique 2\n";
        return false;
    }
    return true;
}


template <typename Container>
bool test_upper_bound(Container c)
{
    using phoenix::arg1;
    using phoenix::arg2;
    using phoenix::upper_bound;

    typename Container::key_type key = c.begin()->first;

    typename Container::iterator lb1 =
        upper_bound(arg1, arg2)(c, key);

    if (lb1 != c.upper_bound(key)) {
	cerr << "Failed " << typeid(Container).name()
             << " test_upper_bound 1\n";
        return false;
    }

    Container const const_c = c;
    key = const_c.begin()->first;

    typename Container::const_iterator lb2 =
	    upper_bound(arg1, arg2)(const_c, key);

    if (lb2 != const_c.upper_bound(key)) {
	cerr << "Failed " << typeid(Container).name()
             << " test_upper_bound 1\n";
        return false;
    }

    return true;
}


template <typename Container>
bool test_value_comp(Container c)
{
    using phoenix::arg1;
    using phoenix::value_comp;

    typename Container::value_compare comp = value_comp(arg1)(c);

    Container const const_c = c;
    comp = value_comp(arg1)(const_c);

    return true;
}


std::deque<int> const build_deque()
{
    std::vector<int> const data = build_vector();
    return std::deque<int>(data.begin(), data.end());
}


std::list<int> const build_list()
{
    std::vector<int> const data = build_vector();
    return std::list<int>(data.begin(), data.end());
}


std::map<int, int> const build_map()
{
    typedef std::map<int, int> int_map;
    typedef std::vector<int> int_vector;

    int_map result;
    int_vector const data = build_vector();
    int_vector::const_iterator it = data.begin();
    int_vector::const_iterator const end = data.end();
    for (; it != end; ++it) {
	    int const value = *it;
	    result[value] = 100 * value;
    }
    return result;
}


std::multimap<int, int> const build_multimap()
{
    typedef std::map<int, int> int_map;
    typedef std::multimap<int, int> int_multimap;
    int_map const data = build_map();
    return int_multimap(data.begin(), data.end());
}


std::vector<int> const init_vector()
{
    typedef std::vector<int> int_vector;
    int const data[] = { -4, -3, -2, -1, 0 };
    int_vector::size_type const data_size = sizeof(data) / sizeof(data[0]);
    return int_vector(data, data + data_size);
}


std::vector<int> const build_vector()
{
    typedef std::vector<int> int_vector;
    static int_vector data = init_vector();
    int_vector::size_type const size = data.size();
    int_vector::iterator it = data.begin();
    int_vector::iterator const end = data.end();
    for (; it != end; ++it)
	    *it += size;
    return data;
}


bool test_deque()
{
    std::deque<int> const data = build_deque();
    bool success = true;
    success = success && test_assign(data);
    success = success && test_assign2(data);
    success = success && test_at(data);
    success = success && test_back(data);
    success = success && test_begin(data);
    success = success && test_clear(data);
    success = success && test_front(data);
    success = success && test_empty(data);
    success = success && test_end(data);
    success = success && test_erase(data);
    success = success && test_get_allocator(data);
    success = success && test_insert(data);
    success = success && test_max_size(data);
    success = success && test_pop_back(data);
    success = success && test_pop_front(data);
    success = success && test_push_back(data);
    success = success && test_push_front(data);
    success = success && test_rbegin(data);
    success = success && test_rend(data);
    success = success && test_resize(data);
    success = success && test_size(data);
    success = success && test_swap(data);
    return success;
}


bool test_list()
{
    std::list<int> const data = build_list();
    bool success = true;
    success = success && test_assign(data);
    success = success && test_assign2(data);
    success = success && test_back(data);
    success = success && test_begin(data);
    success = success && test_clear(data);
    success = success && test_empty(data);
    success = success && test_end(data);
    success = success && test_erase(data);
    success = success && test_front(data);
    success = success && test_get_allocator(data);
    success = success && test_insert(data);
    success = success && test_max_size(data);
    success = success && test_merge(data);
    success = success && test_pop_back(data);
    success = success && test_pop_front(data);
    success = success && test_push_back(data);
    success = success && test_push_front(data);
    success = success && test_rbegin(data);
    success = success && test_remove(data);
    success = success && test_remove_if(data);
    success = success && test_rend(data);
    success = success && test_resize(data);
    success = success && test_reverse(data);
    success = success && test_size(data);
    success = success && test_sort(data);
    success = success && test_splice(data);
    success = success && test_swap(data);
    success = success && test_unique(data);
    return success;
}


bool test_map()
{
    std::map<int, int> const data = build_map();
    bool success = true;
    success = success && test_begin(data);
    success = success && test_clear(data);
    success = success && test_count(data);
    success = success && test_empty(data);
    success = success && test_end(data);
    success = success && test_map_erase(data);
    success = success && test_find(data);
    success = success && test_get_allocator(data);
    success = success && test_map_insert(data);
    success = success && test_key_comp(data);
    success = success && test_lower_bound(data);
    success = success && test_max_size(data);
    success = success && test_rbegin(data);
    success = success && test_rend(data);
    success = success && test_size(data);
    success = success && test_swap(data);
    success = success && test_upper_bound(data);
    success = success && test_value_comp(data);
    return success;
}


bool test_multimap()
{
    std::multimap<int, int> const data = build_multimap();
    bool success = true;
    success = success && test_begin(data);
    success = success && test_clear(data);
    success = success && test_count(data);
    success = success && test_empty(data);
    success = success && test_end(data);
    success = success && test_map_erase(data);
    success = success && test_find(data);
    success = success && test_get_allocator(data);
    success = success && test_multimap_insert(data);
    success = success && test_key_comp(data);
    success = success && test_lower_bound(data);
    success = success && test_max_size(data);
    success = success && test_rbegin(data);
    success = success && test_rend(data);
    success = success && test_size(data);
    success = success && test_swap(data);
    success = success && test_upper_bound(data);
    success = success && test_value_comp(data);
    return success;
}


bool test_vector()
{
    std::vector<int> const data = build_vector();
    bool success = true;
    success = success && test_assign(data);
    success = success && test_assign2(data);
    success = success && test_at(data);
    success = success && test_back(data);
    success = success && test_begin(data);
    success = success && test_capacity(data);
    success = success && test_clear(data);
    success = success && test_end(data);
    success = success && test_empty(data);
    success = success && test_erase(data);
    success = success && test_front(data);
    success = success && test_get_allocator(data);
    success = success && test_insert(data);
    success = success && test_max_size(data);
    success = success && test_pop_back(data);
    success = success && test_push_back(data);
    success = success && test_rbegin(data);
    success = success && test_rend(data);
    success = success && test_reserve(data);
    success = success && test_resize(data);
    success = success && test_size(data);
    success = success && test_swap(data);
    return success;
}

} // namespace anon


int main()
{
    bool success = true;
    success = success && test_deque();
    success = success && test_list();
    success = success && test_map();
    success = success && test_multimap();
    success = success && test_vector();

    return success ? 0 : 1;
}
