/*=============================================================================
    Phoenix v1.2
    Copyright (c) 2004 Angus Leeming

    Use, modification and distribution is subject to the Boost Software
    License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#ifndef PHOENIX_STL_CONTAINER_TRAITS_HPP
#define PHOENIX_STL_CONTAINER_TRAITS_HPP

///////////////////////////////////////////////////////////////////////////////
#include <boost/config.hpp>
#include <boost/mpl/bool.hpp>
#include <boost/mpl/apply_if.hpp>
#include <boost/mpl/identity.hpp>
#include <boost/type_traits/is_same.hpp>
#include <boost/type_traits/is_const.hpp>

///////////////////////////////////////////////////////////////////////////////
//
//  STL container traits
//
//      This file provides metafuctions with which to ascertain
//      whether a class is actually an STL container.
//
//      Four collections of metafuction are provided:
//
//  1   has_allocator_type  has_const_iterator  has_const_reference 
//      has_const_reverse_iterator  has_container_type  has_data_type 
//      has_iterator  has_key_compare  has_key_type  has_reference 
//      has_reverse_iterator  has_size_type  has_value_compare 
//      has_value_type
//
//      derive from "mpl::bool_". Thus
//      "has_const_iterator<std::deque<int> >" derives from "mpl::true_",
//      whilst
//      "has_const_reference<double>" derives from "mpl::false_".
//
//  2   allocator_type_of  const_iterator_of  const_reference_of 
//      const_reverse_iterator_of  container_type_of  data_type_of 
//      iterator_of  key_compare_of  key_type_of  reference_of 
//      reverse_iterator_of  size_type_of  value_compare_of 
//      value_type_of
//
//      provided a typedef "type" to the appropriately named nested type
//      if it exists. If it does not exist then "type" is a typedef for "void".
//
//      Eg, "value_type_of<std::deque<int> >" is "int", whilst
//      "key_type_of<std::vector<double> >" is "void".
//
//  3   cq_iterator_of  cq_reference_of  cq_reverse_iterator_of
//
//      Many of the stl member functions have const and non-const overloaded
//      versions that return distinct types. For example:
//
//      iterator Container::begin();
//      const_iterator Container::begin() const;
//
//      The "Const Qualified" types above will automatcally assign
//      the correct type.
//
//  4   is_std_deque  is_std_list  is_std_map  is_std_multimap 
//      is_std_queue  is_std_vector
//
//      derive from "mpl::bool_" in similar manner to has_const_iterator, etc
//      above.
//
///////////////////////////////////////////////////////////////////////////////

namespace std {

///////////////////////////////////////////////////////////////////////////////
//
// Forward declarations of the supported standard containers.
// No need to pay for stuff you don't use.
//
///////////////////////////////////////////////////////////////////////////////

template <typename ValueT, typename AllocT >
class deque;

template <typename ValueT, typename AllocT >
class list;

template<typename KeyT, typename DataT, typename CompT, typename AllocT>
class map;

template<typename KeyT, typename DataT, typename CompT, typename AllocT>
class multimap;

template<typename ValueT, typename ContainerT>
class queue;

template <typename ValueT, typename AllocT >
class vector;

} // namespace std


///////////////////////////////////////////////////////////////////////////////
namespace phoenix {

namespace stl {
namespace detail {

typedef char One;
typedef struct { char a[2]; } Two;

} // namespace detail

///////////////////////////////////////////////////////////////////////////////
//
// Metafunctions "has_value_type", "has_key_type" etc.
//
//     These metafunctions derive from mpl::true_ if the class
//     defines the named nested type. If it doesn't, then they derive
//     from mpl::false_.
//
//     For example,
//     "has_value_type<std::vector<int> >" derives from mpl::true_.
//     "has_key_type<std::deque<int> >"    derives from mpl::false_.
//
//     These metafunctions are subsequently used to ascertain whether
//     a given class is a std::map etc.
//
// Implementation details.
//
//     * The implementation uses SFINAE and the "sizeof trick" to
//     ascertain whether the nested type exists. It mirrors that
//     described by Vandevoorde & Josuttis, p.106-7.
//     D. Vandevoorde & N.M. Josuttis, "C++ templates: the complete guide",
//     Addison-Wesley, 2003.
//
//     * I use a macro to define structs "has_value_type" etc simply
//     to cut down on the amount of code. The macro is #undef-ed
//     immediately after its final use.
//
//     * g++ 3.2 fails to compile the code if the helper functions
//     "detail::STRUCT_NAME" are placed inside the equivalently-named struct.
//     This failure has nothing to do with the use of a macro to implement
//     the code.
//
///////////////////////////////////////////////////////////////////////////////

#define TYPE_HAS_MEMBER_TYPE(STRUCT_NAME, MEMBER_TYPE)                 \
namespace detail {                                                     \
                                                                       \
/* This fails to work when MEMBER_TYPE is a reference. */              \
/* template <typename TYPE> */                                         \
/* One STRUCT_NAME(typename TYPE::MEMBER_TYPE const *); */             \
template <typename T> One STRUCT_NAME(typename T::MEMBER_TYPE (*)());  \
template <typename T> Two STRUCT_NAME(...);                            \
                                                                       \
} /* namespace detail */                                               \
                                                                       \
template <typename T>                                                  \
struct STRUCT_NAME                                                     \
    : boost::mpl::bool_< sizeof(detail::STRUCT_NAME<T>(0)) == 1 >      \
{}

TYPE_HAS_MEMBER_TYPE(has_allocator_type,   allocator_type);
TYPE_HAS_MEMBER_TYPE(has_const_iterator,   const_iterator);
TYPE_HAS_MEMBER_TYPE(has_const_reference,  const_reference);
TYPE_HAS_MEMBER_TYPE(has_const_reverse_iterator, const_reverse_iterator);
TYPE_HAS_MEMBER_TYPE(has_container_type,   container_type);
TYPE_HAS_MEMBER_TYPE(has_data_type,        data_type);
TYPE_HAS_MEMBER_TYPE(has_iterator,         iterator);
TYPE_HAS_MEMBER_TYPE(has_key_compare,      key_compare);
TYPE_HAS_MEMBER_TYPE(has_key_type,         key_type);
TYPE_HAS_MEMBER_TYPE(has_reference,        reference);
TYPE_HAS_MEMBER_TYPE(has_reverse_iterator, reverse_iterator);
TYPE_HAS_MEMBER_TYPE(has_size_type,        size_type);
TYPE_HAS_MEMBER_TYPE(has_value_compare,    value_compare);
TYPE_HAS_MEMBER_TYPE(has_value_type,       value_type);

#undef TYPE_HAS_MEMBER_TYPE

///////////////////////////////////////////////////////////////////////////////
//
// Metafunctions "value_type_of", "key_type_of" etc.
//
//     These metafunctions define a typedef "type" that
//     returns the nested type if it exists. If not then the typedef
//     returns void.
//
//     For example, "value_type_of<std::vector<int> >::type" is "int"
//     whilst "value_type_of<double>::type" is "void".
//
//     I use a macro to define structs "value_type_of" etc simply
//     to cut down on the amount of code. The macro is #undef-ed
//     immediately after its final use.
//
///////////////////////////////////////////////////////////////////////////////

#define MEMBER_TYPE_OF(STRUCT_NAME, MEMBER_TYPE, HAS_MEMBER_TYPE)  \
template <typename T>                                              \
class STRUCT_NAME {                                                \
    struct member_type { typedef typename T::MEMBER_TYPE type; };  \
public:                                                            \
    typedef typename                                               \
        boost::mpl::apply_if<                                      \
            HAS_MEMBER_TYPE<T>                                     \
          , member_type                                            \
          , boost::mpl::identity<void>                             \
        >::type                                                    \
    type;                                                          \
}

MEMBER_TYPE_OF(allocator_type_of,   allocator_type,  has_allocator_type);
MEMBER_TYPE_OF(const_iterator_of,   const_iterator,  has_const_iterator);
MEMBER_TYPE_OF(const_reference_of,  const_reference, has_const_reference);
MEMBER_TYPE_OF(const_reverse_iterator_of, const_reverse_iterator,
               has_const_reverse_iterator);
MEMBER_TYPE_OF(container_type_of,   container_type,  has_container_type);
MEMBER_TYPE_OF(data_type_of,        data_type,       has_data_type);
MEMBER_TYPE_OF(iterator_of,         iterator,        has_iterator);
MEMBER_TYPE_OF(key_compare_of,      key_compare,     has_key_compare);
MEMBER_TYPE_OF(key_type_of,         key_type,        has_key_type);
MEMBER_TYPE_OF(reference_of,        reference,       has_reference);
MEMBER_TYPE_OF(reverse_iterator_of, reverse_iterator,
               has_reverse_iterator);
MEMBER_TYPE_OF(size_type_of,        size_type,       has_size_type);
MEMBER_TYPE_OF(value_compare_of,    value_compare,   has_value_compare);
MEMBER_TYPE_OF(value_type_of,       value_type,      has_value_type);

#undef MEMBER_TYPE_OF

///////////////////////////////////////////////////////////////////////////////
//
//  Const-Qualified types.
//
//      Many of the stl member functions have const and non-const overloaded
//      versions that return distinct types. For example:
//
//      iterator begin();
//      const_iterator begin() const;
//
//      The three class templates defined below, cq_reference_of,
//      cq_iterator_of and cq_reverse_iterator_of provide a means to extract
//      this return type automatically.
//
//      I use a macro to define the structs simply to cut down on the
//      amount of code. The macro is #undef-ed immediately after its final
//      use.
//
///////////////////////////////////////////////////////////////////////////////

#define CQ_TYPE_OF(STRUCT_NAME, CONST_TYPE, NON_CONST_TYPE) \
template <typename T>                                       \
struct STRUCT_NAME                                          \
{                                                           \
    typedef typename                                        \
        boost::mpl::apply_if<                               \
            boost::is_const<T>                              \
          , CONST_TYPE<T>                                   \
          , NON_CONST_TYPE<T>                               \
        >::type                                             \
    type;                                                   \
}

CQ_TYPE_OF(cq_reference_of, const_reference_of, reference_of);
CQ_TYPE_OF(cq_iterator_of, const_iterator_of, iterator_of);
CQ_TYPE_OF(cq_reverse_iterator_of, const_reverse_iterator_of,
	   reverse_iterator_of);

#undef CQ_TYPE_OF

} // namespace stl

///////////////////////////////////////////////////////////////////////////////
//
//  Is the type a standard container?
//
//      These metafunctions ascertain whether T is the same type
//      as the appropriate STL container generated from it's member types.
//
///////////////////////////////////////////////////////////////////////////////

//////////////////////////////////
template <typename T> struct is_std_deque
   : boost::mpl::false_ {};

template <typename ValueT, typename AllocT>
struct is_std_deque< std::deque<ValueT, AllocT> >
   : boost::mpl::true_ {};

//////////////////////////////////
template <typename T> struct is_std_list
   : boost::mpl::false_ {};

template <typename ValueT, typename AllocT>
struct is_std_list< std::list<ValueT, AllocT> >
   : boost::mpl::true_ {};

//////////////////////////////////
template <typename T> struct is_std_map
   : boost::mpl::false_ {};

template <typename KeyT, typename DataT, typename CompareT, typename AllocT>
struct is_std_map< std::map<KeyT, DataT, CompareT, AllocT> >
   : boost::mpl::true_ {};

//////////////////////////////////
template <typename T> struct is_std_multimap
   : boost::mpl::false_ {};

template <typename KeyT, typename DataT, typename CompareT, typename AllocT>
struct is_std_multimap< std::multimap<KeyT, DataT, CompareT, AllocT> >
   : boost::mpl::true_ {};

//////////////////////////////////
template <typename T> struct is_std_queue
   : boost::mpl::false_ {};

template <typename ValueT, typename ContainerT>
struct is_std_queue< std::queue<ValueT, ContainerT> >
   : boost::mpl::true_ {};

//////////////////////////////////
template <typename T> struct is_std_vector
   : boost::mpl::false_ {};

template <typename ValueT, typename AllocT>
struct is_std_vector< std::vector<ValueT, AllocT> >
   : boost::mpl::true_ {};

} // namespace phoenix

#endif // PHOENIX_STL_CONTAINER_TRAITS_HPP
