/*=============================================================================
    Phoenix v1.2
    Copyright (c) 2004 Angus Leeming

    Use, modification and distribution is subject to the Boost Software
    License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
==============================================================================*/
#ifndef PHOENIX_STL_CONTAINERS_HPP
#define PHOENIX_STL_CONTAINERS_HPP

///////////////////////////////////////////////////////////////////////////////

#if defined PHOENIX_LIMIT && PHOENIX_LIMIT < 5
#undef PHOENIX_LIMIT
#endif

#ifndef PHOENIX_LIMIT
#define PHOENIX_LIMIT 5
#endif

#include "stl_container_traits.hpp"

#include <boost/spirit/phoenix/functions.hpp>
#include <boost/mpl/and.hpp>
#include <boost/mpl/not.hpp>
#include <boost/mpl/or.hpp>
#include <boost/type_traits/is_const.hpp>

///////////////////////////////////////////////////////////////////////////////
namespace phoenix {

///////////////////////////////////////////////////////////////////////////////
//
//  STL container member functions
//
//      Lazy functions for STL container member functions
//
//      These functions provide a mechanism for the lazy evaluation of
//      the public member functions of the STL containers. For an overview
//      of what is meant by 'lazy evaluation', see the comments in
//      operators.hpp and functions.hpp.
//
//      Lazy functions are provided for all of the member functions of the
//      following containers:
//
//      deque  list  map  multimap  vector.
//
//      Indeed, should *your* class have member functions with the
//      same names and signatures as those listed below, then it will
//      also be supported. To summarize, lazy functions are provided for
//      member functions:
//
//      assign  at  back  begin  capacity  clear  count  empty 
//      end  equal_range  erase  find  front  get_allocator  insert 
//      key_comp  lower_bound  max_size  merge  pop_back  pop_front 
//      push_back  push_front  rbegin  remove  remove_if  rend 
//      reserve  resize  reverse  size  sort  splice  swap 
//      unique  upper_bound  value_comp.
//
//      The lazy functions' names are the same as the corresponding
//      member function. Sample usage:
//
//      "Normal" version                 "Lazy" version
//      ----------------                 --------------
//      my_vector.at(5)                  phoenix::at(arg1, 5)
//      my_list.size()                   phoenix::size(arg1)
//      my_vector1.swap(my_vector2)      phoenix::swap(arg1, arg2)
//
//      No support is provided here for lazy versions of
//      operator+=, operator[] etc. Such operators are not specific
//      to STL containers and lazy versions can therefore be found
//      in operators.hpp.
//
///////////////////////////////////////////////////////////////////////////////

namespace stl {

///////////////////////////////////////////////////////////////////////////////
//
//  Lazy member function implementaions.
//
//      The structs below provide the guts of the implementation. Thereafter,
//      the corresponding lazy function itself is simply:
//
//      function<stl::assign> const assign = stl::assign();
//
//      The structs provide a nested "result" class template whose
//      "type" typedef enables the lazy function to ascertain the type
//      to be returned when it is invoked.
//
//      They also provide operator() member functions with signatures
//      corresponding to those of the underlying member function of
//      the STL container.
//
///////////////////////////////////////////////////////////////////////////////

//////////////////////////////////
struct assign {

    template <typename T, typename Arg1 = nil_t, typename Arg2 = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T, typename Arg1, typename Arg2>
    typename result<T>::type
    operator()(T & c, Arg1 const & arg1, Arg2 const & arg2) const
    {
	// The compiler will work out whether to invoke
        //   void Container::assign(size_type const &, value_type const &);
        // or
	//   template <typename It1, typename It2>
	//   void Container::assign(It1 const &, It2 const &);
        return c.assign(arg1, arg2);
    }
};

//////////////////////////////////
struct at {

    template <typename T, typename Index = nil_t>
    struct result {
        typedef typename cq_reference_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c, typename T::size_type const & i) const
    {
        return c.at(i);
    }
};


//////////////////////////////////
struct back {

    template <typename T>
    struct result {
        typedef typename cq_reference_of<T>::type
        type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.back();
    }
};


//////////////////////////////////
struct begin {

    template <typename T>
    struct result {
        typedef typename cq_iterator_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.begin();
    }
};


//////////////////////////////////
struct capacity {

    template <typename T>
    struct result {
        typedef typename size_type_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T const & c) const
    {
        return c.capacity();
    }
};


//////////////////////////////////
struct clear {

    template <typename T>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.clear();
    }
};


//////////////////////////////////
struct count {

    template <typename T, typename Arg1>
    struct result {
        typedef typename size_type_of<T>::type type;
    };

    template <typename T>
    typename result<T, typename T::key_type>::type
    operator()(T & c, typename T::key_type const & key) const
    {
        return c.count(key);
    }
};


//////////////////////////////////
struct empty {

    template <typename T>
    struct result {
        typedef bool type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T const & c) const
    {
        return c.empty();
    }
};


//////////////////////////////////
struct end {

    template <typename T>
    struct result {
        typedef typename cq_iterator_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.end();
    }
};


//////////////////////////////////
struct equal_range {

    template <typename T, typename Arg1>
    class result {
        typedef typename cq_iterator_of<T>::type iterator_type;
    public:
        typedef std::pair<iterator_type, iterator_type> type;
    };

    template <typename T>
    typename result<T, typename T::key_type>::type
    operator()(T & c, typename T::key_type const & key) const
    {
        return c.equal_range(key);
    }
};


//////////////////////////////////
struct erase {

    /*
      This mouthful can differentiate between the generic erase functions
      (Container == deque, list, vector) and that
      specific to the two map-types, std::map and std::multimap.

      iterator Container::erase(iterator where);
      iterator Container::erase(iterator first, iterator last);

      size_type Map::erase(const Key& keyval);
      void Map::erase(iterator where);
      void Map::erase(iterator first, iterator last);

     */
    template <typename T,
              typename Arg1,
              typename Arg2 = nil_t>
    struct result
    {
        typedef
            boost::mpl::apply_if<
                boost::is_same<Arg1, typename iterator_of<T>::type>
              , boost::mpl::identity<void>
              , size_type_of<T>
            >
        map_erase_result;

        typedef typename
            boost::mpl::apply_if<
                boost::mpl::or_<
                    is_std_map<T>
                  , is_std_multimap<T> >
              , map_erase_result
              , iterator_of<T>
            >::type
        type;
    };

    template <typename T, typename Arg1>
    typename result<T, Arg1>::type
    operator()(T & c, Arg1 const & arg1) const
    {
	return c.erase(arg1);
    }

    template <typename T>
    typename result<T, typename T::iterator, typename T::iterator>::type
    operator()(T & c,
               typename T::iterator const & from,
               typename T::iterator const & to) const
    {
        return c.erase(from, to);
    }
};


//////////////////////////////////
struct find {

    template <typename T, typename Arg1 = nil_t>
    struct result {
        typedef typename cq_iterator_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c, typename T::key_type key) const
    {
        return c.find(key);
    }
};


//////////////////////////////////
struct front {

    template <typename T>
    struct result {
        typedef typename cq_reference_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.front();
    }
};


//////////////////////////////////
struct get_allocator {

    template <typename T>
    struct result {
	    typedef typename allocator_type_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T const & c) const
    {
        return c.get_allocator();
    }
};


//////////////////////////////////
struct insert {
    /*
      This mouthful can differentiate between the generic insert functions
      (Container == deque, list, vector) and those specific to the two
      map-types, std::map and std::multimap.

      iterator Container::insert(iterator where, value_type value);
      void Container::insert(iterator where, size_type count, value_type value);
      template <typename ItT>
      void Container::insert(iterator where, ItT first, ItT last);

      pair<iterator, bool> std::map<...>::insert(value_type const &);
      iterator std::multimap<...>::insert(value_type const &);

      template <typename ItT>
      void MAP::insert(ItT first, ItT last);
     */
    template <typename T,
              typename Arg1 = nil_t,
              typename Arg2 = nil_t,
              typename Arg3 = nil_t>
    class result
    {
        struct pair_T_iterator_bool {
            typedef typename std::pair<typename T::iterator, bool> type;
        };

        typedef
            boost::mpl::apply_if<
                is_std_map<T>
              , pair_T_iterator_bool
              , iterator_of<T>
            >
        choice_1;

        typedef
            boost::mpl::apply_if<
                boost::mpl::and_<
                    boost::is_same<Arg3, nil_t>
                  , boost::mpl::not_<boost::is_same<Arg1, Arg2> > >
              , iterator_of<T>
              , boost::mpl::identity<void>
            >
        choice_2;

    public:

        typedef typename
            boost::mpl::apply_if<
		boost::is_same<Arg2, nil_t>
              , choice_1
              , choice_2
            >::type
        type;
    };

    // Handles both
    //     pair<iterator, bool> std::map<...>::insert(value_type const &);
    // and
    //     iterator std::multimap<...>::insert(value_type const &);
    template <typename T>
    typename result<T, typename T::value_type>::type
    operator()(T & c, typename T::value_type value) const
    {
	return c.insert(value);
    }

    template <typename T>
    typename T::iterator
    operator()(
        T & c,
        typename T::iterator const & where,
        typename T::value_type const & value) const
    {
        return c.insert(where, value);
    }

    template <typename T, typename ItT>
    void
    operator()(T & c, ItT first, ItT last) const
    {
	return c.insert(first, last);
    }

    template <typename T>
    void
    operator()(
        T & c,
        typename T::iterator const & where,
        typename T::size_type const & count,
        typename T::value_type const & value) const
    {
        return c.insert(where, count, value);
    }

    template <typename T, typename ItT>
    void
    operator()(
        T & c,
        typename T::iterator const & where,
        ItT const & first,
        ItT const & last) const
    {
        return c.insert(where, first, last);
    }

};


//////////////////////////////////
struct key_comp {

    template <typename T>
    struct result {
        typedef typename key_compare_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T const & c) const
    {
        return c.key_comp();
    }
};


//////////////////////////////////
struct lower_bound {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef typename cq_iterator_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c, typename T::key_type const & key) const
    {
        return c.lower_bound(key);
    }
};


//////////////////////////////////
struct max_size {

    template <typename T>
    struct result {
        typedef typename size_type_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T const & c) const
    {
        return c.max_size();
    }
};


//////////////////////////////////
struct merge {

    template <typename T,
	      typename Arg1 = nil_t,
	      typename Arg2 = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c1, T & c2) const
    {
        return c1.merge(c2);
    }

    template <typename T, typename Predicate>
    typename result<T>::type
    operator()(T & c1, T & c2, Predicate const & predicate) const
    {
        return c1.merge(c2, predicate);
    }
};


//////////////////////////////////
struct pop_back {

    template <typename T>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.pop_back();
    }
};


//////////////////////////////////
struct pop_front {

    template <typename T>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.pop_front();
    }
};


//////////////////////////////////
struct push_back {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c, typename T::value_type const & data) const
    {
        return c.push_back(data);
    }
};


//////////////////////////////////
struct push_front {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c, typename T::value_type const & data) const
    {
        return c.push_front(data);
    }
};


//////////////////////////////////
struct rbegin {

    template <typename T>
    struct result {
        typedef typename cq_reverse_iterator_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.rbegin();
    }
};


//////////////////////////////////
struct remove {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c, typename T::value_type const & value) const
    {
        return c.remove(value);
    }
};


//////////////////////////////////
struct remove_if {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T, typename Predicate>
    typename result<T>::type
    operator()(T & c, Predicate predicate) const
    {
        return c.remove_if(predicate);
    }
};


//////////////////////////////////
struct rend {

    template <typename T>
    struct result {
        typedef typename cq_reverse_iterator_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.rend();
    }
};


//////////////////////////////////
struct reserve {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c, typename T::size_type count) const
    {
        return c.reserve(count);
    }
};


//////////////////////////////////
struct resize {

    template <typename T, typename Arg1 = nil_t, typename Arg2 = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c,
           typename T::size_type new_size) const
    {
        return c.resize(new_size);
    }

    template <typename T>
    typename result<T>::type
    operator()(T & c,
               typename T::size_type new_size,
               typename T::value_type value) const
    {
        return c.resize(new_size, value);
    }
};


//////////////////////////////////
struct reverse {

    template <typename T>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.reverse();
    }
};


//////////////////////////////////
struct size {

    template <typename T>
    struct result {
        typedef typename size_type_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T const & c) const
    {
        return c.size();
    }
};


//////////////////////////////////
struct sort {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.sort();
    }

    template <typename T, typename Predicate>
    typename result<T>::type
    operator()(T & c, Predicate predicate) const
    {
        return c.sort(predicate);
    }
};


//////////////////////////////////
struct splice {

    template <typename Arg1 = nil_t,
	      typename Arg2 = nil_t,
	      typename Arg3 = nil_t,
	      typename Arg4 = nil_t,
	      typename Arg5 = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c,
	       typename T::iterator where,
	       T & right) const
    {
        return c.splice(where, right);
    }

    template <typename T>
    typename result<T>::type
    operator()(T & c,
	       typename T::iterator where,
	       T & right,
	       typename T::iterator first) const
    {
        return c.splice(where, right, first);
    }

    template <typename T>
    typename result<T>::type
    operator()(T & c,
	       typename T::iterator where,
	       T & right,
	       typename T::iterator first,
	       typename T::iterator last) const
    {
        return c.splice(where, right, first, last);
    }
};


//////////////////////////////////
struct swap {

    template <typename T1, typename Arg = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c1, T & c2) const
    {
        return c1.swap(c2);
    }
};


//////////////////////////////////
struct unique {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef void type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c) const
    {
        return c.unique();
    }

    template <typename T, typename Predicate>
    typename result<T>::type
    operator()(T & c, Predicate predicate) const
    {
        return c.unique(predicate);
    }
};


//////////////////////////////////
struct upper_bound {

    template <typename T, typename Arg = nil_t>
    struct result {
        typedef typename cq_iterator_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T & c, typename T::key_type const & key) const
    {
        return c.upper_bound(key);
    }
};


//////////////////////////////////
struct value_comp {

    template <typename T>
    struct result {
        typedef typename value_compare_of<T>::type type;
    };

    template <typename T>
    typename result<T>::type
    operator()(T const & c) const
    {
        return c.value_comp();
    }
};

} // namespace stl


///////////////////////////////////////////////////////////////////////////////
//
//  The lazy functions themselves.
//
///////////////////////////////////////////////////////////////////////////////
function<stl::assign> const assign = stl::assign();
function<stl::at> const at = stl::at();
function<stl::back> const back = stl::back();
function<stl::begin> const begin = stl::begin();
function<stl::capacity> const capacity = stl::capacity();
function<stl::clear> const clear = stl::clear();
function<stl::count> const count = stl::count();
function<stl::empty> const empty = stl::empty();
function<stl::end> const end = stl::end();
function<stl::equal_range> const equal_range = stl::equal_range();
function<stl::erase> const erase = stl::erase();
function<stl::find> const find = stl::find();
function<stl::front> const front = stl::front();
function<stl::get_allocator> const get_allocator = stl::get_allocator();
function<stl::insert> const insert = stl::insert();
function<stl::key_comp> const key_comp = stl::key_comp();
function<stl::lower_bound> const lower_bound = stl::lower_bound();
function<stl::max_size> const max_size = stl::max_size();
function<stl::merge> const merge = stl::merge();
function<stl::pop_back> const pop_back  = stl::pop_back();
function<stl::pop_front> const pop_front = stl::pop_front();
function<stl::push_back> const push_back  = stl::push_back();
function<stl::push_front> const push_front = stl::push_front();
function<stl::rbegin> const rbegin = stl::rbegin();
function<stl::remove> const remove = stl::remove();
function<stl::remove_if> const remove_if = stl::remove_if();
function<stl::rend> const rend = stl::rend();
function<stl::reserve> const reserve = stl::reserve();
function<stl::resize> const resize = stl::resize();
function<stl::reverse> const reverse = stl::reverse();
function<stl::size> const size = stl::size();
function<stl::sort> const sort = stl::sort();
function<stl::splice> const splice = stl::splice();
function<stl::swap> const swap = stl::swap();
function<stl::upper_bound> const upper_bound = stl::upper_bound();
function<stl::unique> const unique = stl::unique();
function<stl::value_comp> const value_comp = stl::value_comp();

} // namespace phoenix

#endif // PHOENIX_STL_CONTAINERS_HPP
